from .feature_selection_network import FeatureSelectionNetwork, gumbel_sigmoid
import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import DataLoader
import tqdm

def select_gumbel_features(X_train, y_train, device="cpu", verbose=False, temperature_decay=0.9999, epochs=300):
    balance = (y_train.sum(axis=0) / len(y_train)).flip(0)

    network = FeatureSelectionNetwork(X_train.shape[1], 16, len(balance)).to(device)

    trainloader = DataLoader(list(zip(X_train, y_train)), batch_size=8)

    criterion = nn.CrossEntropyLoss(weight=balance.to(torch.float32).to(device))
    optimizer = optim.Adam([
        {"params": network.fc1.parameters(), "lr": 3e-6},
        {"params": network.cont.parameters(), "lr": 3e-4},
    ])
    temperature = 2.0

    for epoch in tqdm.trange(epochs, disable=not verbose):
        total = 0
        correct = 0
        epoch_loss = 0
        for data in trainloader:
            X_, y_ = data
            X_, y_ = X_.to(device), y_.to(device).to(torch.float32)

            optimizer.zero_grad()
            output, selected_no = network(X_.float(), temperature=temperature)
            loss = criterion(output, y_.argmax(axis=1)) + selected_no
            loss.backward()
            optimizer.step()
            epoch_loss += loss.item()

            predicted = torch.argmax(output, axis=1)
            real = torch.argmax(y_, axis=1)
            total += y_.size(0)
            correct += (predicted == real).sum().item()
            temperature = temperature * temperature_decay
            
            if epoch > (epochs*2//3):
                temperature = 0
                network.fc1.requires_grad_(False)

    features_sum = torch.zeros(X_train.shape[1], dtype=torch.int32)
    for data in trainloader:
        X_, y_ = data
        X_, y_ = X_.to(device), y_.to(device).to(torch.float32)
        features = network.fc1(X_.float())
        features = gumbel_sigmoid(features, tau=0, hard=True)
        features = features.cpu().detach().numpy()
        features_sum += features.sum(axis=0)
        
    return features_sum.numpy()


from .select_gumbel_features import select_gumbel_features
import torch


class GFSNetwork:
    def __init__(
        self,
        threshold: float = 0.01,
        device: str = "cpu",
        verbose: bool = False,
        temperature_decay: float = 0.9999,
        epochs: int = 300,
    ) -> None:
        """Perform feature selection using GFSNetwork.

        Args:
            threshold (float, optional): Keep features with scores above this percentile score. Defaults to 0.01.
            device (str, optional): Device to use. Defaults to "cpu".
            verbose (bool, optional): Verbosity. Defaults to False.
            temperature_decay (float, optional): Temperature decay. Defaults to 0.9999.
            epochs (int, optional): Number of epochs. Defaults to 300.
        """
        self.threshold = threshold
        self.scores_ = None
        self.device = device
        self.verbose = verbose
        self.temperature_decay = temperature_decay
        self.epochs = epochs

    def fit(self, X, y):
        # cast to torch
        if type(X) != torch.Tensor:
            X = torch.tensor(X)
            y = torch.tensor(y)

        # assert that y is one-hot encoded
        if len(y.shape) == 1:
            y = torch.nn.functional.one_hot(y.to(int))

        self.scores_ = select_gumbel_features(
            X,
            y,
            self.device,
            self.verbose,
            temperature_decay=self.temperature_decay,
            epochs=self.epochs,
        )
        self.ranking_ = self.scores_.argsort()[::-1]
        return self

    @property
    def support_(self):
        assert self.scores_ is not None, "You must call fit before accessing support_"
        perc_score = self.scores_ / self.scores_.max()
        return self.scores_ > perc_score

    def transform(self, X):
        assert self.support_ is not None, "You must call fit before transform"
        return X[:, self.support_]

    def fit_transform(self, X, y):
        self.fit(X, y)
        return self.transform(X)
